/* Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved. */

/******************************************************************************
 *
 * You may not use the identified files except in compliance with the Apache
 * License, Version 2.0 (the "License.")
 *
 * You may obtain a copy of the License at
 * http://www.apache.org/licenses/LICENSE-2.0.
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * The node-oracledb test suite uses 'mocha', 'should' and 'async'. 
 * See LICENSE.md for relevant licenses.
 *
 * NAME
 *   7. autoCommit.js
 *
 * DESCRIPTION
 *   Testing general autoCommit feature.
 *
 * NUMBERING RULE
 *   Test numbers follow this numbering rule:
 *     1  - 20  are reserved for basic functional tests
 *     21 - 50  are reserved for data type supporting tests
 *     51 -     are for other tests 
 * 
 *****************************************************************************/

var oracledb = require('oracledb');
var should   = require('should');
var async    = require('async');
var dbConfig = require('./dbConfig.js');

describe('7. autoCommit.js', function(){
  
  if(dbConfig.externalAuth){
    var credential = { externalAuth: true, connectString: dbConfig.connectString };
  } else {
    var credential = dbConfig;
  }  
  
  var connection = false;
  var anotherConnection = false;
  var script = 
      "BEGIN \
          DECLARE \
              e_table_exists EXCEPTION; \
              PRAGMA EXCEPTION_INIT(e_table_exists, -00942); \
          BEGIN \
              EXECUTE IMMEDIATE ('DROP TABLE oracledb_departments'); \
          EXCEPTION \
              WHEN e_table_exists \
              THEN NULL; \
          END; \
          EXECUTE IMMEDIATE (' \
              CREATE TABLE oracledb_departments ( \
                  department_id NUMBER,  \
                  department_name VARCHAR2(20) \
              ) \
          '); \
      END; ";
  
  beforeEach(function(done){
    oracledb.outFormat = oracledb.OBJECT;
    oracledb.autoCommit = true;
    
    async.series([
      function(callback){
        oracledb.getConnection(credential, function(err, conn){
          if(err) { console.error(err.message); return; }
          connection = conn;
          callback();
        });
      },
      function(callback){
        oracledb.getConnection(credential, function(err, conn){
          if(err) { console.error(err.message); return; }
          anotherConnection = conn;
          callback();
        });
      }, 
      function(callback){
        connection.execute(script, function(err){
          if(err) { console.error(err.message); return; }
          connection.commit( function(err){
            if(err) { console.error(err.message); return; }
            callback();
          });
        });
      }
    ], done);   
  
  
  })
  
  afterEach(function(done){
    oracledb.outFormat = oracledb.ARRAY;
    oracledb.autoCommit = false;
    
    async.series([
      function(callback){
        connection.execute(
          'DROP TABLE oracledb_departments', 
          function(err){
            if(err) { console.error(err.message); return; }
            callback();
          }
        );
      },
      function(callback){
        connection.release( function(err){
          if(err) { console.error(err.message); return; }
          callback();
        });
      },
      function(callback){
        anotherConnection.release( function(err){
          if(err) { console.error(err.message); return; }
          callback();
        });
      }
    ], done);
  })
  
  it('7.1 auto commit takes effect for DML - insert', function(done){
    async.series([
      function(callback){
        connection.execute(
          "INSERT INTO oracledb_departments VALUES (82, 'Security')",
          function(err){
            should.not.exist(err);
            callback();
          }
        );
      },
      function(callback){
        anotherConnection.execute(
          "SELECT department_id FROM oracledb_departments WHERE department_name = 'Security'",
          function(err, result){
            should.not.exist(err);
            should.exist(result);
            // console.log(result);
            result.rows[0].DEPARTMENT_ID.should.eql(82).and.be.a.Number;
            callback();
          }
        );
      }
    ], done);
  })
  
  it('7.2 auto commit takes effect for DML - update', function(done){
    async.series([
      function(callback){
        connection.execute(
          "INSERT INTO oracledb_departments VALUES (82, 'Security')",
          function(err){
            should.not.exist(err);
            callback();
          }
        );
      },
      function(callback){
        connection.execute(
          "UPDATE oracledb_departments SET department_id = 101 WHERE department_name = 'Security'",
          function(err){
            should.not.exist(err);
            callback();
          }
        );
      },
      function(callback){
        anotherConnection.execute(
          "SELECT department_id FROM oracledb_departments WHERE department_name = 'Security'",
          function(err, result){
            should.not.exist(err);
            should.exist(result);
            // console.log(result);
            result.rows[0].DEPARTMENT_ID.should.eql(101).and.be.a.Number;
            callback();
          }
        );
      }
    ], done);
  })
})
